import os
import torch
import torch.nn as nn
from torchvision.models import resnet18

PROJECT_DIR = os.path.dirname(os.path.dirname(os.path.dirname(__file__)))


class Resnet18WS(nn.Module):
    def __init__(self, input_channels=6, pretrained=True):
        super(Resnet18WS, self).__init__()
        self.base = resnet18(pretrained=False)

        if pretrained:
            self.base.load_state_dict(
                torch.load(os.path.join(PROJECT_DIR, 'data/pretrained_model/resnet18-5c106cde.pth'))
            )

        # robust information flow
        self.conv0_1 = nn.Conv2d(input_channels, 64, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), bias=False)
        self.maxPool0_1 = nn.MaxPool2d((2, 2), stride=2)
        self.conv0_2_3 = nn.Sequential(
            nn.Conv2d(64, 64, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), bias=False),
            nn.Conv2d(64, 64, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), bias=False),
        )
        self.maxPool0_2 = nn.MaxPool2d((2, 2), stride=2)

        self.maxPool1 = nn.MaxPool2d((2, 2), stride=2)
        self.maxPool2 = nn.MaxPool2d((2, 2), stride=1)

        self.base.layer2[0].conv1.stride = (1, 1)
        self.base.layer3[0].conv1.stride = (1, 1)
        self.base.layer4[0].conv1.stride = (1, 1)

        self.base.layer2[0].downsample[0].stride = (1, 1)
        del self.base.layer2[0].downsample[1]
        self.base.layer3[0].downsample[0].stride = (1, 1)
        self.base.layer3[0].downsample[0].padding = (0, 0)
        del self.base.layer3[0].downsample[1]
        self.base.layer4[0].downsample[0].stride = (1, 1)
        self.base.layer4[0].downsample[0].padding = (0, 0)
        del self.base.layer4[0].downsample[1]

        # Proposal Feature Alignment (very slight improvement from paper)
        # for i in range(2):
        #     self.base.layer3[i].conv1.dilation = (2, 2)
        #     self.base.layer3[i].conv2.dilation = (2, 2)
        #     self.base.layer4[i].conv1.dilation = (2, 2)
        #     self.base.layer4[i].conv2.dilation = (2, 2)
        # self.base.layer3[0].conv2.dilation = (1, 1)
        # self.base.layer4[0].conv2.dilation = (1, 1)

        self.dim_out = 512
        self.spatial_scale = 1. / 8.

    def forward(self, x):
        x = self.conv0_1(x)
        x = self.maxPool0_1(x)
        x = self.conv0_2_3(x)
        x = self.maxPool0_2(x)
        x = self.base.layer1(x)
        x = self.maxPool1(x)
        x = self.base.layer2(x)
        x = self.maxPool2(x)
        x = self.base.layer3(x)
        x = self.base.layer4(x)

        return x


if __name__ == '__main__':
    resnetws = Resnet18WS(pretrained=False)

    x = torch.randn((1, 6, 640, 480))
    y = resnetws(x)
    print(y.shape)
